# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

cimport numpy as np

cdef struct ShapeletInfo:
    size_t index  # the index of the shapelet sample
    size_t start  # the start position
    size_t length  # the length of the shapelet
    size_t dim  # the dimension of the shapelet
    double mean  # the mean of the shapelet
    double std  # the stanard devision
    void *extra

cdef struct TSDatabase:
    size_t n_samples  # the number of samples
    size_t n_timestep  # the number of timesteps
    size_t n_dims

    double *data  # the data
    size_t sample_stride  # the stride for samples
    size_t timestep_stride  # the `feature` stride
    size_t dim_stride  # the dimension stride

cdef struct Shapelet:
    size_t length
    size_t dim
    double mean
    double std
    int ts_index
    int ts_start
    double *data
    void *extra

cdef class DistanceMeasure:
    cdef size_t n_timestep

    cdef int init_shapelet_info(self, TSDatabase *td, ShapeletInfo *shapelet_info, size_t index, size_t start,
                                size_t length, size_t dim) nogil

    cdef int init_shapelet_ndarray(self, Shapelet *shapelet, np.ndarray arr, size_t dim)

    cdef int init_shapelet(self, Shapelet *shapelet, ShapeletInfo *s, TSDatabase *td) nogil

    cdef int shapelet_matches(self, Shapelet *s, TSDatabase *td, size_t t_index, double threshold, size_t** matches,
                              double** distances, size_t *n_matches) nogil except -1

    cdef double shapelet_distance(self, Shapelet *s, TSDatabase *td, size_t t_index, size_t *return_index= *) nogil

    cdef double shapelet_info_distance(self, ShapeletInfo *s, TSDatabase *td, size_t t_index) nogil

    cdef void shapelet_info_distances(self, ShapeletInfo *s, TSDatabase *td, size_t *samples, double *distances,
                                      size_t n_samples) nogil

cdef class ScaledDistanceMeasure(DistanceMeasure):
    pass

cdef TSDatabase ts_database_new(np.ndarray X)

cdef void shapelet_info_init(ShapeletInfo *s) nogil

cdef void shapelet_info_free(ShapeletInfo *shapelet_info) nogil

cdef int shapelet_init(Shapelet *shapelet, size_t dim, size_t length, double mean, double std) nogil

cdef void shapelet_free(Shapelet *shapelet) nogil



# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten


cdef double scaled_euclidean_distance(size_t s_offset,
                                      size_t s_stride,
                                      size_t s_length,
                                      double s_mean,
                                      double s_std,
                                      double *S,
                                      size_t t_offset,
                                      size_t t_stride,
                                      size_t t_length,
                                      double *T,
                                      double *X_buffer,
                                      size_t *index) nogil

cdef double euclidean_distance(size_t s_offset,
                               size_t s_stride,
                               size_t s_length,
                               double *S,
                               size_t t_offset,
                               size_t t_stride,
                               size_t t_length,
                               double *T,
                               size_t *index) nogil



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

import numpy as np
cimport numpy as np

from libc.stdlib cimport malloc

from libc.math cimport sqrt
from libc.math cimport NAN

import wildboar._dtw_distance
import wildboar._euclidean_distance

from libc.stdlib cimport free
from sklearn.utils import check_array
from ._distance cimport Shapelet
from ._distance cimport DistanceMeasure
from ._distance cimport ts_database_new
from ._distance cimport TSDatabase

_DISTANCE_MEASURE = {
    'euclidean': wildboar._euclidean_distance.EuclideanDistance,
    'scaled_euclidean': wildboar._euclidean_distance.ScaledEuclideanDistance,
    'scaled_dtw': wildboar._dtw_distance.ScaledDtwDistance,
}

cdef int shapelet_init(Shapelet *shapelet, size_t dim, size_t length, double mean, double std) nogil:
    shapelet[0].dim = dim
    shapelet[0].length = length
    shapelet[0].mean = mean
    shapelet[0].std = std
    shapelet[0].data = <double*> malloc(sizeof(double) * length)
    if shapelet[0].data == NULL:
        return -1

cdef void shapelet_free(Shapelet *shapelet) nogil:
    if shapelet != NULL and shapelet[0].data != NULL:
        free(shapelet[0].data)

cdef void shapelet_info_init(ShapeletInfo *s) nogil:
    """Initialize  a shapelet info struct """
    s[0].start = 0
    s[0].length = 0
    s[0].dim = 0
    s[0].mean = NAN
    s[0].std = NAN
    s[0].index = 0

cdef void shapelet_info_free(ShapeletInfo *shapelet_info) nogil:
    """Free the `extra` payload of a shapelet info if needed """
    if shapelet_info[0].extra != NULL:
        free(shapelet_info[0].extra)
        shapelet_info[0].extra = NULL

cdef int _shapelet_info_update_statistics(ShapeletInfo *s, const TSDatabase *t_ptr) nogil:
    """Update the mean and standard deviation of a shapelet info struct """
    cdef TSDatabase t = t_ptr[0]
    cdef size_t shapelet_offset = (s.index * t.sample_stride +
                                   s.dim * t.dim_stride +
                                   s.start * t.timestep_stride)
    cdef double ex = 0
    cdef double ex2 = 0
    cdef size_t i
    for i in range(s.length):
        current_value = t.data[shapelet_offset + i * t.timestep_stride]
        ex += current_value
        ex2 += current_value ** 2

    s[0].mean = ex / s.length
    s[0].std = sqrt(ex2 / s.length - s[0].mean * s[0].mean)
    return 0

cdef TSDatabase ts_database_new(np.ndarray data):
    """Construct a new time series database from a ndarray """
    if data.ndim < 2 or data.ndim > 3:
        raise ValueError("ndim {0} < 2 or {0} > 3".format(data.ndim))

    cdef TSDatabase sd
    sd.n_samples = <size_t> data.shape[0]
    sd.n_timestep = <size_t> data.shape[data.ndim - 1]
    sd.data = <double*> data.data
    sd.sample_stride = <size_t> data.strides[0] / <size_t> data.itemsize
    sd.timestep_stride = (<size_t> data.strides[data.ndim - 1] /
                          <size_t> data.itemsize)

    if data.ndim == 3:
        sd.n_dims = <size_t> data.shape[data.ndim - 2]
        sd.dim_stride = (<size_t> data.strides[data.ndim - 2] /
                         <size_t> data.itemsize)
    else:
        sd.n_dims = 1
        sd.dim_stride = 0

    return sd

cdef class DistanceMeasure:
    """A distance measure can compute the distance between time series and
    shapelets """

    def __cinit__(self, size_t n_timestep, *args, **kvargs):
        """ Initialize a new distance measure

        :param n_timesteps: the (maximum) number of timepoints in a timeseries

        :param *args: optimal arguments for subclasses

        :param **kvargs: optional arguments for subclasses
        """
        self.n_timestep = n_timestep

    def __reduce__(self):
        return self.__class__, (self.n_timestep,)

    cdef void shapelet_info_distances(
            self, ShapeletInfo *s, TSDatabase *td, size_t *samples,
            double *distances, size_t n_samples) nogil:
        """ Compute the distance between the shapelet `s` in `td` and all
        samples in `samples`

        :param s: information about the shapelet
        :param td: the time series database
        :param samples: array of length `n_samples` samples to compute
        the distance to
        :param distances: array to store the distances. The the length
        of distances >= `n_samples`, the `i = 0,...,n_samples`
        position stores the distance between the i:th sample (in
        `samples`) and `s` [out param]
        :param n_samples: the number of samples 
        """
        cdef size_t p
        for p in range(n_samples):
            distances[p] = self.shapelet_info_distance(s, td, samples[p])

    cdef int init_shapelet_info(
            self, TSDatabase *_td, ShapeletInfo *shapelet_info, size_t index, size_t start,
            size_t length, size_t dim) nogil:
        """Return a information about a shapelet

        :param _td: shapelet database
        :param shapelet_info: [out param] 
        :param index: the index of the sample in `td`
        :param start: the start position of the subsequence
        :param length: the length of the subsequence
        :param dim: the dimension of the subsequence
        :return non-negative on success
        """
        shapelet_info[0].index = index
        shapelet_info[0].dim = dim
        shapelet_info[0].start = start
        shapelet_info[0].length = length
        shapelet_info[0].mean = NAN
        shapelet_info[0].std = NAN
        shapelet_info[0].extra = NULL
        return 0

    cdef int init_shapelet_ndarray(self, Shapelet *shapelet, np.ndarray arr, size_t dim):
        shapelet[0].dim = dim
        shapelet[0].length = arr.shape[0]
        shapelet[0].mean = NAN
        shapelet[0].std = NAN
        shapelet[0].data = <double*> malloc(shapelet[0].length * sizeof(double))
        if shapelet[0].data == NULL:
            return -1

        cdef size_t i
        for i in range(shapelet[0].length):
            shapelet[0].data[i] = arr[i]
        return 0

    cdef int init_shapelet(self, Shapelet *shapelet, ShapeletInfo *s_ptr, TSDatabase *td_ptr) nogil:
        cdef ShapeletInfo s = s_ptr[0]
        cdef TSDatabase td = td_ptr[0]
        shapelet_init(shapelet, s.dim, s.length, s.mean, s.std)
        shapelet[0].ts_start = s.start
        shapelet[0].ts_index = s.index
        cdef double *data = shapelet[0].data
        cdef size_t shapelet_offset = (s.index * td.sample_stride +
                                       s.start * td.timestep_stride +
                                       s.dim * td.dim_stride)

        cdef size_t i
        cdef size_t p

        for i in range(s.length):
            p = shapelet_offset + td.timestep_stride * i
            data[i] = td.data[p]

        return 0

    cdef double shapelet_info_distance(
            self, ShapeletInfo *si, TSDatabase *td_ptr, size_t t_index) nogil:
        """Return the distance between `s` and the sample specified by the
        index `t_index` in `td`. Implemented by subclasses.

        :param si: shapelet information

        :param td_ptr: the time series database

        :param t_index: the index of the time series
        """
        with gil:
            raise NotImplementedError()

    cdef double shapelet_distance(
            self, Shapelet *s_ptr, TSDatabase *td_ptr, size_t t_index,
            size_t *return_index=NULL) nogil:
        """Return the distance between `s` and the sample specified by
        `t_index` in `td` setting the index of the best matching
        position to `return_index` unless `return_index == NULL`

        :param s_ptr: the shapelet

        :param td_ptr: the time series database

        :param t_index: the sample index

        :param return_index: (out) the index of the best matching position
        """
        with gil:
            raise NotImplementedError()

    cdef int shapelet_matches(
            self, Shapelet *s_ptr, TSDatabase *td_ptr, size_t t_index,
            double threshold, size_t** matches, double** distances,
            size_t *n_matches) nogil except -1:
        """Compute the matches for `s` in the sample `t_index` in `td` where
        the distance threshold is below `threshold`, storing the
        matching starting positions in `matches` and distance (<
        `threshold`) in `distances` with `n_matches` storing the
        number of successful matches.

        Note:

        - `matches` will be allocated and must be freed by the caller
        - `distances` will be allocated and must be freed by the caller

        :param s_ptr: the shapelet

        :param td_ptr: the time series database

        :param t_index: the sample

        :param threshold: the minimum distance to consider a match

        :param matches: (out) array of matching locations

        :param distances: (out) array of distance at the matching
        location (< `threshold`)

        :param n_matches: (out) the number of matches
        """
        with gil:
            raise NotImplementedError()

cdef class ScaledDistanceMeasure(DistanceMeasure):
    """Distance measure that uses computes the distance on mean and
    variance standardized shapelets"""

    cdef int init_shapelet_ndarray(self, Shapelet *shapelet, np.ndarray arr, size_t dim):
        cdef int err = DistanceMeasure.init_shapelet_ndarray(self, shapelet, arr, dim)
        if err == -1:
            return -1
        shapelet[0].mean = np.mean(arr)
        shapelet[0].std = np.std(arr)
        return 0

    cdef int init_shapelet_info(self, TSDatabase *td, ShapeletInfo *shapelet_info, size_t index, size_t start,
                                size_t length, size_t dim) nogil:
        DistanceMeasure.init_shapelet_info(self, td, shapelet_info, index, start, length, dim)
        _shapelet_info_update_statistics(shapelet_info, td)
        return -1

# cdef class FuncDistanceMeasure(DistanceMeasure):
#     """ Wrapper for python function """
#     cdef object func
#     cdef np.ndarray[ndim=1, dtype=np.float] sbuffer
#     cdef np.ndarray[ndim=1, dtype=np.float] tbuffer

#     def __cinit__(self, size_t n_timesteps, object func):
#         self.func = func
#         self.sbuffer = np.empty([n_timesteps], dtype=np.float32)
#         self.tbuffer = np.empty([n_timesteps], dtype=np.float32)

#     cdef double shapelet_info_distance(
#             self, ShapeletInfo s, TSDatabase td, size_t t_index) nogil:
#         with gil:
#             raise NotImplementedError()

#     cdef double shapelet_distance(
#             self, Shapelet s, TSDatabase td, size_t t_index,
#             size_t* return_index=NULL) nogil:
#         with gil:
#             raise NotImplementedError()

#     cdef int shapelet_matches(
#             self, Shapelet s, TSDatabase td, size_t t_index,
#             double threhold, size_t** matches,  double** distances,
#             size_t* n_matches) nogil except -1:
#         with gil:
#             raise NotImplementedError()    

def _validate_shapelet(shapelet):
    cdef np.ndarray s = check_array(
        shapelet, ensure_2d=False, dtype=np.float64, order="c")
    if s.ndim > 1:
        raise ValueError("only 1d shapelets allowed")

    if not s.flags.contiguous:
        s = np.ascontiguousarray(s, dtype=np.float64)
    return s

def _validate_data(data):
    cdef np.ndarray x = check_array(
        data, ensure_2d=False, allow_nd=True, dtype=np.float64, order="c")
    if x.ndim == 1:
        x = x.reshape(-1, x.shape[0])

    if not x.flags.contiguous:
        x = np.ascontiguousarray(x, dtype=np.float64)
    return x

def _check_sample(sample, n_samples):
    if sample < 0 or sample >= n_samples:
        raise ValueError("illegal sample {}".format(sample))

def _check_dim(dim, ndims):
    if dim < 0 or dim >= ndims:
        raise ValueError("illegal dimension {}".format(dim))

cdef np.ndarray _new_match_array(size_t *matches, size_t n_matches):
    if n_matches > 0:
        match_array = np.empty(n_matches, dtype=np.intp)
        for i in range(n_matches):
            match_array[i] = matches[i]
        return match_array
    else:
        return None

cdef np.ndarray _new_distance_array(
        double *distances, size_t n_matches):
    if n_matches > 0:
        dist_array = np.empty(n_matches, dtype=np.float64)
        for i in range(n_matches):
            dist_array[i] = distances[i]
        return dist_array
    else:
        return None

def distance(shapelet, data, dim=0, sample=None, metric="euclidean", metric_params=None, return_index=False):
    """Computes the minimum distance between `s` and the samples in `x`

    :param shapelet: the subsequence `array_like`

    :param data: the samples `[n_samples, n_timesteps]` or
                 `[n_samples, n_dims, n_timesteps]`

    :param dim: the time series dimension to search (default: 0)

    :param sample: the samples to compare to `int` or `array_like` or
                   `None`; if `sample` is `None`, return the distance
                   to all samples in data. Note that if, `n_samples`
                   is 1 a scalar is returned; otherwise a array is
                   returned.

    :param metric: the distance measure

    :param return_index: if `true` return the index of the best
                         match. If there are many equally good
                         best matches, the first is returned.

    :returns: `float`, `(float, int)`, `float [n_samples]` or `(float
              [n_samples], int [n_samples]` depending on input

    """
    cdef np.ndarray s = _validate_shapelet(shapelet)
    cdef np.ndarray x = _validate_data(data)
    if sample is None:
        if x.shape[0] == 1:
            sample = 0
        else:
            sample = np.arange(x.shape[0])

    cdef TSDatabase sd = ts_database_new(x)

    _check_dim(dim, sd.n_dims)
    cdef double min_dist
    cdef size_t min_index

    cdef double mean = 0
    cdef double std = 0

    if metric_params is None:
        metric_params = {}

    cdef DistanceMeasure distance_measure = _DISTANCE_MEASURE[metric](
        sd.n_timestep, **metric_params)

    cdef Shapelet shape
    distance_measure.init_shapelet_ndarray(&shape, s, dim)
    if isinstance(sample, int):
        min_dist = distance_measure.shapelet_distance(
            &shape, &sd, sample, &min_index)

        if return_index:
            return min_dist, min_index
        else:
            return min_dist
    else:  # assume an `array_like` object for `samples`
        samples = check_array(sample, ensure_2d=False, dtype=np.int)
        dist = []
        ind = []
        for i in samples:
            min_dist = distance_measure.shapelet_distance(&shape, &sd, i, &min_index)

            dist.append(min_dist)
            ind.append(min_index)

        if return_index:
            return np.array(dist), np.array(ind)
        else:
            return np.array(dist)

def matches(shapelet, X, threshold, dim=0, sample=None, metric="euclidean", metric_params=None, return_distance=False):
    """Return the positions in data (one array per `sample`) where
    `shapelet` is closer than `threshold`.

    :param shapelet: the subsequence `array_like`

    :param X: the samples `[n_samples, n_timestep]` or `[n_sample,
              n_dim, n_timestep]`

    :param threshold: the maximum threshold for match

    :param dim: the time series dimension to search (default: 0)

    :param sample: the samples to compare to `int` or `array_like` or
                   `None`.  If `None` compare to all. (default:
                   `None`)

    :param metric: the distance measure

    :param metric_params: additional parameters to the metric function
                          (optional, dict, default: None)

    :returns: `[n_matches]`, or `[[n_matches], ... n_samples]`

    """
    cdef np.ndarray s = _validate_shapelet(shapelet)
    cdef np.ndarray x = _validate_data(X)
    _check_dim(dim, x.ndim)
    if sample is None:
        if x.shape[0] == 1:
            sample = 0
        else:
            sample = np.arange(x.shape[0])

    cdef TSDatabase sd = ts_database_new(x)

    cdef size_t *matches
    cdef double *distances
    cdef size_t n_matches

    if metric_params is None:
        metric_params = {}

    cdef DistanceMeasure distance_measure = _DISTANCE_MEASURE[metric](
        sd.n_timestep, **metric_params)
    cdef Shapelet shape
    distance_measure.init_shapelet_ndarray(&shape, s, dim)

    cdef size_t i
    if isinstance(sample, int):
        _check_sample(sample, sd.n_samples)
        distance_measure.shapelet_matches(
            &shape, &sd, sample, threshold, &matches, &distances, &n_matches)

        match_array = _new_match_array(matches, n_matches)
        distance_array = _new_distance_array(distances, n_matches)
        free(distances)
        free(matches)

        if return_distance:
            return distance_array, match_array
        else:
            return match_array
    else:
        samples = check_array(sample, ensure_2d=False, dtype=np.int)
        match_list = []
        distance_list = []
        for i in samples:
            _check_sample(i, sd.n_samples)
            distance_measure.shapelet_matches(&shape, &sd, i, threshold, &matches, &distances, &n_matches)
            match_array = _new_match_array(matches, n_matches)
            distance_array = _new_distance_array(distances, n_matches)
            match_list.append(match_array)
            distance_list.append(distance_array)
            free(matches)
            free(distances)

        if return_distance:
            return distance_list, match_list
        else:
            return match_list



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

# This implementation is heavily inspired by the UCRSuite.
#
# References
#
#  - Rakthanmanon, et al., Searching and Mining Trillions of Time
#    Series Subsequences under Dynamic Time Warping (2012)
#  - http://www.cs.ucr.edu/~eamonn/UCRsuite.html
cimport numpy as np

from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.math cimport INFINITY
from libc.math cimport sqrt
from libc.math cimport floor

from ._distance cimport TSDatabase

from ._distance cimport ShapeletInfo
from ._distance cimport ScaledDistanceMeasure
from ._distance cimport Shapelet

cdef void deque_init(Deque *c, size_t capacity) nogil:
    c[0].capacity = capacity
    c[0].size = 0
    c[0].queue = <size_t*> malloc(sizeof(size_t) * capacity)
    c[0].front = 0
    c[0].back = capacity - 1

cdef void deque_reset(Deque *c) nogil:
    c[0].size = 0
    c[0].front = 0
    c[0].back = c[0].capacity - 1

cdef void deque_destroy(Deque *c) nogil:
    free(c[0].queue)

cdef void deque_push_back(Deque *c, size_t v) nogil:
    c[0].queue[c[0].back] = v
    c[0].back -= 1
    if c[0].back < 0:
        c[0].back = c[0].capacity - 1

    c[0].size += 1

cdef void deque_pop_front(Deque *c) nogil:
    c[0].front -= 1
    if c[0].front < 0:
        c[0].front = c[0].capacity - 1
    c[0].size -= 1

cdef void deque_pop_back(Deque *c) nogil:
    c[0].back = (c[0].back + 1) % c[0].capacity
    c[0].size -= 1

cdef size_t deque_front(Deque *c) nogil:
    cdef int tmp = c[0].front - 1
    if tmp < 0:
        tmp = c[0].capacity - 1
    return c[0].queue[tmp]

cdef size_t deque_back(Deque *c) nogil:
    cdef int tmp = (c[0].back + 1) % c[0].capacity
    return c[0].queue[tmp]

cdef bint deque_empty(Deque *c) nogil:
    return c[0].size == 0

cdef size_t deque_size(Deque *c) nogil:
    return c[0].size

cdef void find_min_max(size_t offset, size_t stride, size_t length,
                       double *T, size_t r, double *lower, double *upper,
                       Deque *dl, Deque *du) nogil:
    cdef size_t i
    cdef size_t k

    cdef double current, prev

    deque_reset(du)
    deque_reset(dl)

    deque_push_back(du, 0)
    deque_push_back(dl, 0)

    for i in range(1, length):
        if i > r:
            k = i - r - 1
            upper[k] = T[offset + stride * deque_front(du)]
            lower[k] = T[offset + stride * deque_front(dl)]

        current = T[offset + stride * i]
        prev = T[offset + stride * (i - 1)]
        if current > prev:
            deque_pop_back(du)
            while (not deque_empty(du) and
                   current > T[offset + stride * deque_back(du)]):
                deque_pop_back(du)
        else:
            deque_pop_back(dl)
            while (not deque_empty(dl) and
                   current < T[offset + stride * deque_back(dl)]):
                deque_pop_back(dl)

        deque_push_back(du, i)
        deque_push_back(dl, i)

        if i == 2 * r + 1 + deque_front(du):
            deque_pop_front(du)
        elif i == 2 * r + 1 + deque_front(dl):
            deque_pop_front(dl)

    for i in range(length, length + r + 1):
        upper[i - r - 1] = T[offset + stride * deque_front(du)]
        lower[i - r - 1] = T[offset + stride * deque_front(dl)]

        if i - deque_front(du) >= 2 * r + 1:
            deque_pop_front(du)
        if i - deque_front(dl) >= 2 * r + 1:
            deque_pop_front(dl)

cdef inline double dist(double x, double y) nogil:
    cdef double s = x - y
    return s * s

cdef double constant_lower_bound(size_t s_offset, size_t s_stride, double *S,
                                 double s_mean, double s_std, size_t t_offset,
                                 size_t t_stride, double *T, double t_mean,
                                 double t_std, size_t length,
                                 double best_dist) nogil:
    cdef double t_x0, t_y0, s_x0, s_y0
    cdef double t_x1, ty1, s_x1, s_y1
    cdef double t_x2, t_y2, s_x2, s_y2
    cdef double distance, min_dist

    # first and last in T
    t_x0 = (T[t_offset] - t_mean) / t_std
    t_y0 = (T[t_offset + t_stride * (length - 1)] - t_mean) / t_std

    # first and last in S
    s_x0 = (S[s_offset] - s_mean) / s_std
    s_y0 = (S[s_offset + s_stride * (length - 1)] - s_mean) / s_std

    min_dist = dist(t_x0, s_x0) + dist(t_y0, s_y0)
    if min_dist >= best_dist:
        return min_dist

    t_x1 = (T[t_offset + t_stride * 1] - t_mean) / t_std
    s_x1 = (S[s_offset + s_stride * 1] - s_mean) / s_std
    min_dist += min(
        min(dist(t_x1, s_x0), dist(t_x0, s_x1)),
        dist(t_x1, s_x1))

    if min_dist >= best_dist:
        return min_dist

    t_y1 = (T[t_offset + t_stride * (length - 2)] - t_mean) / t_std
    s_y1 = (S[s_offset + s_stride * (length - 2)] - s_mean) / s_std
    min_dist += min(
        min(dist(t_y1, s_y1), dist(t_y0, s_y1)),
        dist(t_y1, s_y1))

    if min_dist >= best_dist:
        return min_dist

    t_x2 = (T[t_offset + t_stride * 2] - t_mean) / t_std
    s_x2 = (S[s_offset + s_stride * 2] - s_mean) / s_std
    min_dist += min(min(dist(t_x0, s_x2),
                        min(dist(t_x1, s_x2),
                            dist(t_x2, s_x2)),
                        dist(t_x2, s_x1)),
                    dist(t_x2, s_x0))

    if min_dist >= best_dist:
        return min_dist

    t_y2 = (T[t_offset + t_stride * (length - 3)] - t_mean) / t_std
    s_y2 = (S[s_offset + s_stride * (length - 3)] - s_mean) / s_std

    min_dist += min(min(dist(t_y0, s_y2),
                        min(dist(t_y1, s_y2),
                            dist(t_y2, s_y2)),
                        dist(t_y2, s_y1)),
                    dist(t_y2, s_y0))

    return min_dist

cdef double cumulative_bound(size_t offset, size_t stride, size_t length,
                             double mean, double std, double *T,
                             double lu_mean, double lu_std, double *lower,
                             double *upper, double *cb, double best_so_far) nogil:
    cdef double min_dist = 0
    cdef double x, d, us, ls
    cdef size_t i

    for i in range(0, length):
        if min_dist >= best_so_far:
            break

        x = (T[offset + stride * i] - mean) / std
        us = (upper[i] - lu_mean) / lu_std
        ls = (lower[i] - lu_mean) / lu_std
        if x > us:
            d = dist(x, us)
        elif x < ls:
            d = dist(x, ls)
        else:
            d = 0

        min_dist += d
        cb[i] = d
    return min_dist

cdef inline double inner_dtw(size_t s_offset, size_t s_stride, int s_length,
                             double s_mean, double s_std, double *S,
                             double mean, double std, size_t x_offset,
                             double *X_buffer, int r, double *cb,
                             double *cost, double *cost_prev,
                             double min_dist) nogil:
    cdef int i = 0
    cdef int j = 0
    cdef int k = 0

    cdef double x
    cdef double y
    cdef double z
    cdef double min_cost, distance

    cdef double *cost_tmp
    for i in range(0, 2 * r + 1):
        cost[i] = INFINITY
        cost_prev[i] = INFINITY

    for i in range(0, s_length):
        k = max(0, r - i)
        min_cost = INFINITY
        for j in range(max(0, i - r), min(s_length, i + r + 1)):
            if i == 0 and j == 0:
                min_cost = dist((S[s_offset] - s_mean) / s_std,
                                (X_buffer[x_offset] - mean) / std)
                cost[k] = min_cost
            else:
                if j - 1 < 0 or k - 1 < 0:
                    y = INFINITY
                else:
                    y = cost[k - 1]

                if i - 1 < 0 or k + 1 > 2 * r:
                    x = INFINITY
                else:
                    x = cost_prev[k + 1]

                if i - 1 < 0 or j - 1 < 0:
                    z = INFINITY
                else:
                    z = cost_prev[k]

                distance = dist((S[s_offset + s_stride * i] - s_mean) / s_std,
                                (X_buffer[x_offset + j] - mean) / std)
                cost[k] = min(min(x, y), z) + distance
                if cost[k] < min_cost:
                    min_cost = cost[k]

            k += 1

        if i + r < s_length - 1 and min_cost + cb[i + r + 1] >= min_dist:
            return min_cost + cb[i + r + 1]

        cost_tmp = cost
        cost = cost_prev
        cost_prev = cost_tmp
    return cost_prev[k - 1]

cdef double scaled_dtw_distance(size_t s_offset,
                                size_t s_stride,
                                size_t s_length,
                                double s_mean,
                                double s_std,
                                double *S,
                                size_t t_offset,
                                size_t t_stride,
                                size_t t_length,
                                double *T,
                                size_t r,
                                double *X_buffer,
                                double *cost,
                                double *cost_prev,
                                double *s_lower,
                                double *s_upper,
                                double *t_lower,
                                double *t_upper,
                                double *cb,
                                double *cb_1,
                                double *cb_2,
                                size_t *index) nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef double lb_kim
    cdef double lb_k
    cdef double lb_k2

    cdef double ex = 0
    cdef double ex2 = 0

    cdef size_t i
    cdef size_t j
    cdef size_t k
    cdef size_t I
    cdef size_t buffer_pos

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value

        if i >= s_length - 1:
            j = (i + 1) % s_length
            I = i - (s_length - 1)
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            lb_kim = constant_lower_bound(s_offset, s_stride, S,
                                          s_mean, s_std, j, 1, X_buffer,
                                          mean, std, s_length, min_dist)

            if lb_kim < min_dist:
                lb_k = cumulative_bound(j, 1, s_length, mean, std, X_buffer,
                                        s_mean, s_std, s_lower, s_upper,
                                        cb_1, min_dist)
                if lb_k < min_dist:
                    lb_k2 = cumulative_bound(
                        s_offset, s_stride, s_length, s_mean, s_std, S,
                        mean, std, t_lower + I, t_upper + I, cb_2, min_dist)

                    if lb_k2 < min_dist:
                        if lb_k > lb_k2:
                            cb[s_length - 1] = cb_1[s_length - 1]
                            for k in range(s_length - 2, -1, -1):
                                cb[k] = cb[k + 1] + cb_1[k]
                        else:
                            cb[s_length - 1] = cb_2[s_length - 1]
                            for k in range(s_length - 2, -1, -1):
                                cb[k] = cb[k + 1] + cb_2[k]
                        dist = inner_dtw(
                            s_offset, s_stride, s_length, s_mean,
                            s_std, S, mean, std, j, X_buffer, r, cb,
                            cost, cost_prev, min_dist)

                        if dist < min_dist:
                            if index != NULL:
                                index[0] = (i + 1) - s_length
                            min_dist = dist

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return sqrt(min_dist)

cdef inline size_t _compute_warp_width(size_t length, double r) nogil:
    if r == 1:
        return length - 1
    if r < 1:
        return <size_t> floor(length * r)
    else:
        return <size_t> floor(r)

cdef class ScaledDtwDistance(ScaledDistanceMeasure):
    cdef double *X_buffer
    cdef double *lower
    cdef double *upper
    cdef double *cost
    cdef double *cost_prev
    cdef double *cb
    cdef double *cb_1
    cdef double *cb_2

    cdef Deque du
    cdef Deque dl

    cdef size_t max_warp_width
    cdef double r

    def __cinit__(self, size_t n_timestep, double r=0):
        super().__init__(n_timestep)
        if r < 0:
            raise ValueError("illegal warp width")
        self.r = r
        self.max_warp_width = _compute_warp_width(n_timestep, self.r)
        self.X_buffer = <double*> malloc(sizeof(double) * n_timestep * 2)
        self.lower = <double*> malloc(sizeof(double) * n_timestep)
        self.upper = <double*> malloc(sizeof(double) * n_timestep)
        self.cost = <double*> malloc(sizeof(double) * 2 * self.max_warp_width + 1)
        self.cost_prev = <double*> malloc(sizeof(double) * 2 * self.max_warp_width + 1)
        self.cb = <double*> malloc(sizeof(double) * n_timestep)
        self.cb_1 = <double*> malloc(sizeof(double) * n_timestep)
        self.cb_2 = <double*> malloc(sizeof(double) * n_timestep)

        if (self.X_buffer == NULL or
                self.lower == NULL or
                self.upper == NULL or
                self.cost == NULL or
                self.cost_prev == NULL or
                self.cb == NULL or
                self.cb_1 == NULL or
                self.cb_2 == NULL):
            raise MemoryError()

        deque_init(&self.dl, 2 * self.max_warp_width + 2)
        deque_init(&self.du, 2 * self.max_warp_width + 2)

    def __dealloc__(self):
        free(self.X_buffer)
        free(self.lower)
        free(self.upper)
        free(self.cost)
        free(self.cost_prev)
        free(self.cb)
        free(self.cb_1)
        free(self.cb_2)

    cdef int init_shapelet_info(self, TSDatabase *td_ptr, ShapeletInfo *shapelet_info, size_t index, size_t start,
                                size_t length, size_t dim) nogil:
        cdef TSDatabase td = td_ptr[0]
        ScaledDistanceMeasure.init_shapelet_info(self, td_ptr, shapelet_info, index, start, length, dim)

        cdef DtwExtra *dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        cdef size_t shapelet_offset = (index * td.sample_stride +
                                       start * td.timestep_stride +
                                       dim * td.dim_stride)
        cdef size_t warp_width = _compute_warp_width(length, self.r)
        find_min_max(shapelet_offset, td.timestep_stride, length, td.data,
                     warp_width, dtw_extra[0].lower, dtw_extra[0].upper,
                     &self.dl, &self.du)

        shapelet_info[0].extra = dtw_extra
        return 0

    cdef int init_shapelet(self, Shapelet *shapelet, ShapeletInfo *si_ptr, TSDatabase *td_ptr) nogil:
        cdef int err = ScaledDistanceMeasure.init_shapelet(self, shapelet, si_ptr, td_ptr)
        if err == -1:
            return -1

        cdef TSDatabase td = td_ptr[0]
        cdef ShapeletInfo shapelet_info = si_ptr[0]

        cdef DtwExtra *dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        cdef size_t length = shapelet[0].length
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        cdef size_t warp_width = _compute_warp_width(length, self.r)
        find_min_max(0, 1, length, shapelet[0].data,
                     warp_width, dtw_extra[0].lower, dtw_extra[0].upper,
                     &self.dl, &self.du)
        shapelet[0].extra = dtw_extra
        return 0

    cdef int init_shapelet_ndarray(self, Shapelet *shapelet, np.ndarray arr, size_t dim):
        cdef int err = ScaledDistanceMeasure.init_shapelet_ndarray(self, shapelet, arr, dim)
        if err == -1:
            return -1
        cdef size_t length = shapelet[0].length
        cdef DtwExtra *dtw_extra = <DtwExtra*> malloc(sizeof(DtwExtra))
        dtw_extra[0].lower = <double*> malloc(sizeof(double) * length)
        dtw_extra[0].upper = <double*> malloc(sizeof(double) * length)

        cdef size_t warp_width = _compute_warp_width(length, self.r)
        find_min_max(0, 1, length, shapelet[0].data,
                     warp_width, dtw_extra[0].lower, dtw_extra[0].upper,
                     &self.dl, &self.du)
        shapelet[0].extra = dtw_extra
        return 0

    cdef double shapelet_distance(self, Shapelet *s_ptr, TSDatabase *td_ptr, size_t t_index,
                                  size_t *return_index=NULL) nogil:
        cdef TSDatabase td = td_ptr[0]
        cdef Shapelet s = s_ptr[0]
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        cdef double *s_lower
        cdef double *s_upper
        cdef DtwExtra *extra
        cdef size_t warp_width = _compute_warp_width(s.length, self.r)

        if s.extra != NULL:
            extra = <DtwExtra*> s.extra
            s_lower = extra[0].lower
            s_upper = extra[0].upper
        else:
            s_lower = <double*> malloc(sizeof(double) * s.length)
            s_upper = <double*> malloc(sizeof(double) * s.length)

            find_min_max(0, 1, s.length, s.data, warp_width, s_lower, s_upper,
                         &self.dl, &self.du)

        find_min_max(sample_offset, td.timestep_stride, td.n_timestep,
                     td.data, warp_width, self.lower, self.upper,
                     &self.dl, &self.du)

        cdef double distance = scaled_dtw_distance(0,
                                                   1,
                                                   s.length,
                                                   s.mean,
                                                   s.std,
                                                   s.data,
                                                   sample_offset,
                                                   td.timestep_stride,
                                                   td.n_timestep,
                                                   td.data,
                                                   warp_width,
                                                   self.X_buffer,
                                                   self.cost,
                                                   self.cost_prev,
                                                   s_lower,
                                                   s_upper,
                                                   self.lower,
                                                   self.upper,
                                                   self.cb,
                                                   self.cb_1,
                                                   self.cb_2,
                                                   return_index)
        if s.extra == NULL:
            free(s_lower)
            free(s_upper)

        return distance

    cdef double shapelet_info_distance(self, ShapeletInfo *s_ptr, TSDatabase *td_ptr, size_t t_index) nogil:
        cdef TSDatabase td = td_ptr[0]
        cdef ShapeletInfo s = s_ptr[0]
        cdef size_t sample_offset = (t_index * td.sample_stride +
                                     s.dim * td.dim_stride)
        cdef size_t shapelet_offset = (s.index * td.sample_stride +
                                       s.dim * td.dim_stride +
                                       s.start * td.timestep_stride)

        cdef size_t warp_width = _compute_warp_width(s.length, self.r)

        cdef DtwExtra *dtw_extra = <DtwExtra*> s.extra
        find_min_max(sample_offset, td.timestep_stride, td.n_timestep,
                     td.data, warp_width, self.lower, self.upper, &self.dl,
                     &self.du)
        return scaled_dtw_distance(shapelet_offset,
                                   td.timestep_stride,
                                   s.length,
                                   s.mean,
                                   s.std,
                                   td.data,
                                   sample_offset,
                                   td.timestep_stride,
                                   td.n_timestep,
                                   td.data,
                                   warp_width,
                                   self.X_buffer,
                                   self.cost,
                                   self.cost_prev,
                                   dtw_extra[0].lower,
                                   dtw_extra[0].upper,
                                   self.lower,
                                   self.upper,
                                   self.cb,
                                   self.cb_1,
                                   self.cb_2,
                                   NULL)



# -*- coding: utf-8 -*-

import sys
import os
import ast

from setuptools import setup
from setuptools.extension import Extension


def build_ext(*args, **kwargs):
    import Cython.Build
    return Cython.Build.build_ext(*args, **kwargs)


libname = "wildboar"
build_type = "optimized"

SHORTDESC = "wildboar is the fundamental package for time series classification with Python"

DESC = """
It provides:

 * Shapelet tree classification and regression
 * Random shapelet forest classification and regression
 * Fast dynamic time warning searching
 * Fast euclidean distance searching

The package is provided under the GPLv3 license.
"""

datadirs = ("test",)
dataexts = (".py", ".pyx", ".pxd", ".c", ".cpp", ".h", ".sh", ".lyx", ".tex",
            ".txt", ".pdf")

standard_docs = ["README", "LICENSE", "TODO", "CHANGELOG", "AUTHORS"]
standard_doc_exts = [".md", ".rst", ".txt", "", ".org"]

if sys.version_info < (3, 4):
    sys.exit('Sorry, Python < 3.4 is not supported')

extra_compile_args_math_optimized = [
    '-march=native',
    '-O2',
    '-msse',
    '-msse2',
    '-mfma',
    '-mfpmath=sse',
]
extra_compile_args_math_debug = [
    '-march=native',
    '-O0',
    '-g',
]

extra_link_args_math_optimized = []
extra_link_args_math_debug = []

extra_compile_args_nonmath_optimized = ['-O2']
extra_compile_args_nonmath_debug = ['-O0', '-g']
extra_link_args_nonmath_optimized = []
extra_link_args_nonmath_debug = []

openmp_compile_args = ['-fopenmp']
openmp_link_args = ['-fopenmp']


# Lazy loading
class np_include_dirs(str):
    def __str__(self):
        import numpy as np
        return np.get_include()


my_include_dirs = [np_include_dirs()]

if build_type == 'optimized':
    my_extra_compile_args_math = extra_compile_args_math_optimized
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_optimized
    my_extra_link_args_math = extra_link_args_math_optimized
    my_extra_link_args_nonmath = extra_link_args_nonmath_optimized
    my_debug = False
    print("build configuration selected: optimized")
elif build_type == 'debug':
    my_extra_compile_args_math = extra_compile_args_math_debug
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_debug
    my_extra_link_args_math = extra_link_args_math_debug
    my_extra_link_args_nonmath = extra_link_args_nonmath_debug
    my_debug = True
    print("build configuration selected: debug")
else:
    raise ValueError(
        "Unknown build configuration '%s'; valid: 'optimized', 'debug'" %
        (build_type))


def declare_cython_extension(extName,
                             use_math=False,
                             use_openmp=False,
                             include_dirs=None,
                             extra_lib=None):
    extPath = extName.replace(".", os.path.sep) + ".pyx"
    if use_math:
        compile_args = list(my_extra_compile_args_math)  # copy
        link_args = list(my_extra_link_args_math)
        import platform
        if platform.system() == "Windows":
            libraries = None
        else:
            libraries = ["m"]
    else:
        compile_args = list(my_extra_compile_args_nonmath)
        link_args = list(my_extra_link_args_nonmath)
        libraries = None

    if use_openmp:
        compile_args.insert(0, openmp_compile_args)
        link_args.insert(0, openmp_link_args)

    if extra_lib is not None:
        if libraries is None:
            libraries = []
        for lib in extra_lib:
            libraries.append(lib)

    return Extension(
        extName, [extPath],
        extra_compile_args=compile_args,
        extra_link_args=link_args,
        include_dirs=include_dirs,
        libraries=libraries)


datafiles = []


def getext(filename):
    return os.path.splitext(filename)[1]


for datadir in datadirs:
    datafiles.extend(
        [(root,
          [os.path.join(root, f) for f in files if getext(f) in dataexts])
         for root, dirs, files in os.walk(datadir)])

detected_docs = []
for docname in standard_docs:
    for ext in standard_doc_exts:
        filename = "".join((docname, ext))
        if os.path.isfile(filename):
            detected_docs.append(filename)
datafiles.append(('.', detected_docs))

init_py_path = os.path.join(libname, '__init__.py')
version = '0.0.unknown'
try:
    with open(init_py_path) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            print(
                "WARNING: Version information not found"
                " in '%s', using placeholder '%s'" % (init_py_path, version),
                file=sys.stderr)
except FileNotFoundError:
    print(
        "WARNING: Could not find file '%s',"
        "using placeholder version information '%s'" % (init_py_path, version),
        file=sys.stderr)

ext_module_utils = declare_cython_extension(
    "wildboar._utils",
    use_math=True,
    use_openmp=False,
    include_dirs=my_include_dirs)

ext_module_distance = declare_cython_extension(
    "wildboar._distance",
    use_math=True,
    use_openmp=False,
    include_dirs=my_include_dirs)

ext_module_euclidean_distance = declare_cython_extension(
    "wildboar._euclidean_distance",
    use_math=True,
    use_openmp=False,
    include_dirs=my_include_dirs)

ext_module_dtw_distance = declare_cython_extension(
    "wildboar._dtw_distance",
    use_math=True,
    use_openmp=False,
    include_dirs=my_include_dirs)

ext_module_impurity = declare_cython_extension(
    "wildboar._impurity",
    use_math=True,
    use_openmp=False,
    include_dirs=my_include_dirs)

ext_module_tree_builder = declare_cython_extension(
    "wildboar._tree_builder",
    use_math=True,
    use_openmp=False,
    include_dirs=my_include_dirs)

cython_ext_modules = [
    ext_module_utils,
    ext_module_distance,
    ext_module_euclidean_distance,
    ext_module_dtw_distance,
    ext_module_impurity,
    ext_module_tree_builder,
]

setup(
    name="wildboar",
    version=version,
    author="Isak Samsten",
    author_email="isak@samsten.se",
    url="https://github.com/isakkarlsson/wildboar",
    description=SHORTDESC,
    long_description=DESC,
    license="GPLv3",
    #    platforms=["Linux"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries",
    ],
    setup_requires=[
        'cython>=0.28',
        'numpy>=1.14.2',
        'setuptools>=18.0',
    ],
    install_requires=["scikit-learn"],
    python_requires=">=3.4.0",
    provides=["wildboar"],
    keywords=["machine learning", "time series distance"],
    ext_modules=cython_ext_modules,
    packages=["wildboar"],
    package_data={
        'wildboar': ['*.pxd', '*.pyx', '*.c'],
    },
    zip_safe=False,
    #    cmdclass={'build_ext': build_ext},

    # Custom data files not inside a Python package
    data_files=datafiles,
)
